using UnityEngine;
using System.Collections;

// Note: the tile sprites are picked using BlockType cast to int as index to an array
// (see: BoardTile)
public enum BlockType : int {
    L,
    J,
    T,
    Z,
    S,
    I,
    O
}

// Block contains information about a block's positions in Point2 space given its type, orientation, and position
public struct Block {
    enum BlockOrientation {
        Up,
        Left,
        Down,
        Right
    }

    BlockType type;
    BlockOrientation orientation;
    Point2 position;
    
    // The Point2 positions corresponding to the rotation & block type adjusted by the block's position
    // (for tetrominoes this will always have Length == 4)
    public Point2[] GridPositions { get; private set; }

    // Starting orientation is ALWAYS up
    public Block( BlockType type, Point2 position ) {
        this.type = type;
        this.orientation = BlockOrientation.Up;
        this.position = position;
        this.GridPositions = BaseGridPositions( type, BlockOrientation.Up );
        this.GridPositions.ShiftBy( position );
    }

    public Block( Block toCopy ) {
        this.type = toCopy.type;
        this.orientation = toCopy.orientation;
        this.position = toCopy.position;
        this.GridPositions = new Point2[4];
        for ( int i = 0; i < 4; i++ ) {
            this.GridPositions[i] = toCopy.GridPositions[i];
        }
    }

    // Returns a new block corresponding to rotating 'this' to the right without modifying its position
    public Block RotatedRight() {
        var newBlock = new Block( this );
        switch ( orientation ) {
            case BlockOrientation.Up:
                newBlock.orientation = BlockOrientation.Right;
                break;
            case BlockOrientation.Right:
                newBlock.orientation = BlockOrientation.Down;
                break;
            case BlockOrientation.Down:
                newBlock.orientation = BlockOrientation.Left;
                break;
            case BlockOrientation.Left:
                newBlock.orientation = BlockOrientation.Up;
                break;
        }
        newBlock.GridPositions = BaseGridPositions( newBlock.type, newBlock.orientation );
        newBlock.GridPositions.ShiftBy( position );
        return newBlock;
    }

    // Returns a new block corresponding to rotating 'this' to the left without modifying its position
    public Block RotatedLeft() {
        var newBlock = new Block( this );
        switch ( orientation ) {
            case BlockOrientation.Up:
                newBlock.orientation = BlockOrientation.Left;
                break;
            case BlockOrientation.Left:
                newBlock.orientation = BlockOrientation.Down;
                break;
            case BlockOrientation.Down:
                newBlock.orientation = BlockOrientation.Right;
                break;
            case BlockOrientation.Right:
                newBlock.orientation = BlockOrientation.Up;
                break;
        }
        newBlock.GridPositions = BaseGridPositions( newBlock.type, newBlock.orientation );
        newBlock.GridPositions.ShiftBy( position );
        return newBlock;
    }

    // Returns a new block corresponding to shifting 'this' by 'offset' without modifying its rotation
    public Block ShiftedBy( Point2 offset ) {
        var newBlock = new Block( this );
        newBlock.position += offset;
        newBlock.GridPositions.ShiftBy(offset);
        return newBlock;
    }

    // Returns the array of local Point2s corresponding to a block type and its rotation
    static Point2[] BaseGridPositions( BlockType type, BlockOrientation orientation ) {

        bool horizontal = (orientation == BlockOrientation.Up || orientation == BlockOrientation.Down);

        var positions = new Point2[4];
        positions[0] = Point2.Zero;  // Origin point is always the first point

        switch ( type ) {
            case BlockType.L:
                /*   Up      Left     Down    Right
                 * ______[3] [3][2] [2](0)[1] [1]___
                 * [1](0)[2] ___(0) [3]______ (0)___
                 *           ___[1]           [2][3]
                 */
                switch ( orientation ) {
                    case BlockOrientation.Up:
                        positions[1] = new Point2( -1, 0 );
                        positions[2] = new Point2( 1, 0 );
                        positions[3] = new Point2( 1, 1 );
                        break;
                    case BlockOrientation.Left:
                        positions[1] = new Point2( 0, -1 );
                        positions[2] = new Point2( 0, 1 );
                        positions[3] = new Point2( -1, 1 );
                        break;
                    case BlockOrientation.Down:
                        positions[1] = new Point2( 1, 0 );
                        positions[2] = new Point2( -1, 0 );
                        positions[3] = new Point2( -1, -1 );
                        break;
                    case BlockOrientation.Right:
                        positions[1] = new Point2( 0, 1 );
                        positions[2] = new Point2( 0, -1 );
                        positions[3] = new Point2( 1, -1 );
                        break;
                }
                break;

            case BlockType.J:
                /*   Up      Left     Down    Right
                 * [3]______ ___[1] [1](0)[2] [2][3]
                 * [2](0)[1] ___(0) ______[3] (0)___
                 *           [3][2]           [1]___
                 */
                switch ( orientation ) {
                    case BlockOrientation.Up:
                        positions[1] = new Point2( 1, 0 );
                        positions[2] = new Point2( -1, 0 );
                        positions[3] = new Point2( -1, 1 );
                        break;
                    case BlockOrientation.Left:
                        positions[1] = new Point2( 0, 1 );
                        positions[2] = new Point2( 0, -1 );
                        positions[3] = new Point2( -1, -1 );
                        break;
                    case BlockOrientation.Down:
                        positions[1] = new Point2( -1, 0 );
                        positions[2] = new Point2( 1, 0 );
                        positions[3] = new Point2( 1, -1 );
                        break;
                    case BlockOrientation.Right:
                        positions[1] = new Point2( 0, -1 );
                        positions[2] = new Point2( 0, 1 );
                        positions[3] = new Point2( 1, 1 );
                        break;
                }
                break;

            case BlockType.T:
                /*    Up      Left    Down    Right
                 * ___[2]___ ___[3] [3](0){1] [1]___
                 * [1](0)[3] [2](0) ___[2]___ (0)[2]
                 *           ___[1]           [3]___
                 */
                switch ( orientation ) {
                    case BlockOrientation.Up:
                        positions[1] = new Point2( -1, 0 );
                        positions[2] = new Point2( 0, 1 );
                        positions[3] = new Point2( 1, 0 );
                        break;
                    case BlockOrientation.Left:
                        positions[1] = new Point2( 0, -1 );
                        positions[2] = new Point2( -1, 0 );
                        positions[3] = new Point2( 0, 1 );
                        break;
                    case BlockOrientation.Down:
                        positions[1] = new Point2( 1, 0 );
                        positions[2] = new Point2( 0, -1 );
                        positions[3] = new Point2( -1, 0 );
                        break;
                    case BlockOrientation.Right:
                        positions[1] = new Point2( 0, 1 );
                        positions[2] = new Point2( 1, 0 );
                        positions[3] = new Point2( 0, -1 );
                        break;
                }
                break;

            case BlockType.Z:
                /*   Horiz.   Vert.
                 * [2][1]___ ___[2]
                 * ___(0){3] (0)[1]
                 *           [3]___
                 */
                if ( horizontal ) {
                    positions[1] = new Point2( 0, 1 );
                    positions[2] = new Point2( -1, 1 );
                    positions[3] = new Point2( 1, 0 );
                } else {
                    positions[1] = new Point2( 1, 0 );
                    positions[2] = new Point2( 1, 1 );
                    positions[3] = new Point2( 0, -1 );
                }
                break;

            case BlockType.S:
                /*   Horiz.   Vert.
                 * ___[1]{2] [3]___
                 * [3](0)___ (0)[1]
                 *           ___[2]
                 */
                if ( horizontal ) {
                    positions[1] = new Point2( 0, 1 );
                    positions[2] = new Point2( 1, 1 );
                    positions[3] = new Point2( -1, 0 );
                } else {
                    positions[1] = new Point2( 1, 0 );
                    positions[2] = new Point2( 1, -1 );
                    positions[3] = new Point2( 0, 1 );
                }
                break;

            case BlockType.I:
                /*    Horiz.   Vert.
                 * [1](0)[2][3] [1]
                 *              (0)
                 *              [2]
                 *              [3]
                 */
                if ( horizontal ) {
                    positions[1] = new Point2( -1, 0 );
                    positions[2] = new Point2( 1, 0 );
                    positions[3] = new Point2( 2, 0 );
                } else {
                    positions[1] = new Point2( 0, 1 );
                    positions[2] = new Point2( 0, -1 );
                    positions[3] = new Point2( 0, -2 );
                }
                break;

            case BlockType.O:
                /* [1][2]
                 * (0)[3]
                 */
                positions[1] = new Point2( 0, 1 );
                positions[2] = new Point2( 1, 1 );
                positions[3] = new Point2( 1, 0 );
                break;
        }
        return positions;
    }
}
