using UnityEngine;
using UnityEngine.UI;
using System;
using System.Collections;

public class BoardManager : MonoBehaviour {

    // Note: each tile in the grid is 1x1 world units
    const int BOARD_WIDTH = 10;
    const int BOARD_HEIGHT = 22;
    const int BLOCK_SPAWN_HEIGHT = 20;  // Top two squares will be hidden, but allow blocks to be rotated/placed

    static Vector3 BOTTOM_LEFT_POSITION {
        get { return new Vector3( 0f - (BOARD_WIDTH / 2f), 0f - (BOARD_HEIGHT / 2f), 0f ); }
    }

    // Called with current score as the parameter when the new block cannot be drawn in its initial position
    public event Action<int> OnGameEnd;

    public bool Playing { get; private set; }

    [SerializeField] GameObject boardTilePrefab;
    [SerializeField] PreviewBoard previewBoard;
    [SerializeField] Text scoreText;

    BoardTile[,] grid = new BoardTile[BOARD_WIDTH, BOARD_HEIGHT];

    Block block;
    // The new block is spawned immediately when CurrentBlockType is set
    BlockType _currentBlockType;
    BlockType CurrentBlockType {
        get { return _currentBlockType; }
        set {
            _currentBlockType = value;
            block = new Block( _currentBlockType, new Point2( (BOARD_WIDTH / 2) - 1, BLOCK_SPAWN_HEIGHT ) );
            // Shift is applied immediately to move the block into a fully visible position and check if it's valid
            if ( !ShiftBlockDown() ) {
                EndGame();  // If the block can't draw, the game is over
            }
        }
    }
    // The preview window is updated immediately when NextBlockType is set
    BlockType _nextBlockType;
    BlockType NextBlockType {
        get { return _nextBlockType; }
        set {
            _nextBlockType = value;
            previewBoard.SetBlockPreview( _nextBlockType );
        }
    }

    int _score;
    int Score {
        get { return _score; }
        set {
            _score = value;
            string scoreFmt = "00000000";  // 8 seems like enough digits
            scoreText.text = _score.ToString( scoreFmt );
        }
    }

    void Awake() {
        UnityEngine.Random.seed = (int)System.DateTime.Now.ToBinary();
        // Set up the board
        for ( int x = 0; x < BOARD_WIDTH; x += 1 ) {
            for ( int y = 0; y < BOARD_HEIGHT; y += 1 ) {
                var newObj = Instantiate( boardTilePrefab ) as GameObject;
                newObj.transform.SetParent( this.transform );
                newObj.transform.localPosition = BOTTOM_LEFT_POSITION + new Vector3( x, y, 0 );
                grid[x, y] = newObj.GetComponent<BoardTile>();
            }
        }
        Playing = false;
    }

    public void StartGame() {
        Playing = true;
        Score = 0;
        // Clear the board
        foreach ( var tile in grid ) {
            tile.TileIndex = -1;
        }
        CurrentBlockType = GetRandomBlockType();
        NextBlockType = GetRandomBlockType();
    }

    public void EndGame() {
        Playing = false;
        OnGameEnd( Score );
    }

    // Returns false on a frame where it had to spawn a new block, returns true otherwise
    public bool StepFrame() {
        var shiftedSuccessfully = ShiftBlockDown();

        if ( !shiftedSuccessfully ) {
            FinalizeBlockPlacement();
        }
        return shiftedSuccessfully;
    }

    // Note: these Rotate and Shift functions return a bool indicating success
    // This could allow the game to play a "bump" sound when trying to move the block into a wall (for example)

    public bool RotateBlockRight() {
        var rotatedBlock = block.RotatedRight();
        return TryToMoveBlock( rotatedBlock );
    }

    public bool RotateBlockLeft() {
        var rotatedBlock = block.RotatedLeft();
        return TryToMoveBlock( rotatedBlock );
    }

    public bool ShiftBlockDown() {
        var downshiftedBlock = block.ShiftedBy( new Point2( 0, -1 ) );
        return TryToMoveBlock( downshiftedBlock );
    }

    public bool ShiftBlockRight() {
        var rightshiftedBlock = block.ShiftedBy( new Point2( 1, 0 ) );
        return TryToMoveBlock( rightshiftedBlock );
    }

    public bool ShiftBlockLeft() {
        var leftshiftedBlock = block.ShiftedBy( new Point2( -1, 0 ) );
        return TryToMoveBlock( leftshiftedBlock );
    }

    // Shift down until the block can't move any more
    public void DropBlockDownward() {
        while ( ShiftBlockDown() ) { }
        FinalizeBlockPlacement();
    }

    // Check to see if the shifted block will be valid. If it is, the shift is applied
    bool TryToMoveBlock( Block shiftedBlock ) {
        if ( !Playing ) {
            return false;
        }

        var nextGridPos = shiftedBlock.GridPositions;
        foreach ( var pos in nextGridPos ) {
            if ( !BoardPositionIsValid( pos ) ||
                ( TileAt( pos ).IsOccupied && !block.GridPositions.Contains( pos ) ) ) {
                return false;
            }
        }
        // No conflicts; execute shift
        ClearCurrentBlockLocation();
        block = shiftedBlock;
        DrawCurrentBlockLocation();
        return true;
    }

    // Check to see if a given Point2 is in bounds for the board
    bool BoardPositionIsValid( Point2 point ) {
        return point.x >= 0 && point.x < BOARD_WIDTH
            && point.y >= 0 && point.y < BOARD_HEIGHT;
    }

    BoardTile TileAt( Point2 point ) {
        return grid[point.x, point.y];
    }

    BlockType GetRandomBlockType() {
        return (BlockType)UnityEngine.Random.Range( 0, 7 );
    }

    // Clear full rows, shift the blocks down, and return the number of rows cleared
    int ClearCompletedRows() {
        int numRowsCleared = 0;
        int outerRow = 0;
        while ( outerRow < BOARD_HEIGHT )
        {
            bool rowIsFilled = true;
            for ( int column = 0; column < BOARD_WIDTH; column++ ) {
                if ( !grid[column, outerRow].IsOccupied )
                {
                    rowIsFilled = false;
                    break;
                }
            }

            if ( rowIsFilled ) {
                numRowsCleared++;
                // Shift down the rows above
                for ( int row = outerRow; row < BOARD_HEIGHT; row++ ) {
                    for ( int column = 0; column < BOARD_WIDTH; column++ ) {
                        // Take the new TileIndex from the row above (or make it -1 if at the top row)
                        var newIndex = ( row + 1 < BOARD_HEIGHT ) ? grid[column, row + 1].TileIndex : -1;
                        grid[column, row].TileIndex = newIndex;
                    }
                }
                continue;  // Skip outerRow++ to check the newly fallen row
            }
            
            outerRow++;
        }
        return numRowsCleared;
    }

    void ClearCurrentBlockLocation() {
        foreach ( var pos in block.GridPositions ) {
            grid[pos.x, pos.y].TileIndex = -1;
        }
    }

    void DrawCurrentBlockLocation() {
        foreach ( var pos in block.GridPositions ) {
            grid[pos.x, pos.y].TileIndex = (int)CurrentBlockType;
        }
    }

    // Clear completed rows and spawn a new block
    void FinalizeBlockPlacement() {
        var rowsCleared = ClearCompletedRows();
        if ( rowsCleared > 0 ) {
            switch ( rowsCleared ) {
                case 4:
                    Score += 800;
                    break;
                case 3:
                    Score += 500;
                    break;
                case 2:
                    Score += 300;
                    break;
                default:
                    Score += 100;
                    break;
            }
        }

        CurrentBlockType = NextBlockType;
        NextBlockType = GetRandomBlockType();
    }

    // Wire outlines to help with positioning
    void OnDrawGizmos() {
        Gizmos.color = Color.blue;
        Gizmos.DrawWireCube(
            transform.localPosition - new Vector3( 0f, (BOARD_HEIGHT - BLOCK_SPAWN_HEIGHT) / 2f, 0f ),
            new Vector3( 1 + BOARD_WIDTH, 1 + BOARD_HEIGHT - (BOARD_HEIGHT - BLOCK_SPAWN_HEIGHT), 1f )
            );
        Gizmos.color = Color.red;
        Gizmos.DrawWireCube(
            transform.localPosition,
            new Vector3( 1 + BOARD_WIDTH, 1 + BOARD_HEIGHT, 1f )
            );
    }
}
