using UnityEngine;
using System.Collections;

[RequireComponent( typeof( BoardManager ) )]
public class GameController : MonoBehaviour {
    [SerializeField] GameObject titleScreenObject;
    [SerializeField] GameObject gameOverObject;
    [SerializeField] float stepPeriod;  // Amount of time between board steps (the block falling)
    float nextStepTime;

    // 'shifts' refer to the basic left/right/down movement of the block (shifting it one unit on the grid)
    [SerializeField] float shiftPeriod;  // Amount of time between shifts while the button is held down
    float nextShiftTime;

    // Store a reference to the shift function while its corresponding button is held
    delegate bool ShiftDelegate();
    ShiftDelegate ExecuteShift;

    BoardManager board;

    void Awake() {
        board = GetComponent<BoardManager>();
        board.OnGameEnd += GameOver;
        ExecuteShift = null;
        gameOverObject.SetActive( false );
        titleScreenObject.SetActive( true );
    }

    public void NewGame() {
        titleScreenObject.SetActive( false );
        gameOverObject.SetActive( false );
        board.StartGame();
        nextStepTime = Time.time + stepPeriod;
    }

    void GameOver( int score ) {
        gameOverObject.SetActive( true );
    }

    void Update() {
        if ( board.Playing ) {
            // The shift function is nullified to indicate release
            if ( ExecuteShift != null && Time.time > nextShiftTime ) {
                nextShiftTime += shiftPeriod;
                ExecuteShift();
            }

            if ( Time.time > nextStepTime ) {
                nextStepTime += stepPeriod;
                board.StepFrame();
            }
        }
    }

    // Allow holding functionality for each of the basic shift operations (left, right, down)
    void HandleShiftInput( bool starting, ShiftDelegate shiftFunction ) {
        if ( starting ) {
            HandleShiftInputStart( shiftFunction );
        } else {
            HandleShiftInputEnd( shiftFunction );
        }
    }

    void HandleShiftInputStart( ShiftDelegate shiftFunction ) {
        shiftFunction();
        ExecuteShift = shiftFunction;
        // Wait double the amount of time for the first press
        nextShiftTime = Time.time + (shiftPeriod * 2);
    }

    void HandleShiftInputEnd( ShiftDelegate shiftFunction ) {
        // Only stop our own shift
        if ( ExecuteShift == shiftFunction ) {
            ExecuteShift = null;
        }
    }

    // Public functions for button input

    // Holdable functions are sent with true when pressed and false when released
    // (see the EventTrigger components)

    public void InputDown( bool starting ) {
        HandleShiftInput( starting, board.ShiftBlockDown );
    }

    public void InputRight( bool starting ) {
        HandleShiftInput( starting, board.ShiftBlockRight );
    }

    public void InputLeft( bool starting ) {
        HandleShiftInput( starting, board.ShiftBlockLeft );
    }

    // Don't do the holding pattern with the drop or spins

    public void InputDrop() {
        board.DropBlockDownward();
    }

    public void InputSpinRight() {
        board.RotateBlockRight();
    }

    public void InputSpinLeft() {
        board.RotateBlockLeft();
    }
}
