// Point2 offers a convenient way to think about & do algebra on 2D integer-based grid
public struct Point2 {
    public int x, y;

    public Point2( int x, int y ) {
        this.x = x;
        this.y = y;
    }

    public static Point2 Zero { get { return new Point2( 0, 0 ); } }

    public override bool Equals( object obj ) {
        return obj is Point2 && this == (Point2)obj;
    }

    public override int GetHashCode() {
        return x.GetHashCode() ^ y.GetHashCode();
    }

    public override string ToString() {
        return string.Format( "({0}, {1})", this.x, this.y );
    }

    public static bool operator ==( Point2 left, Point2 right ) {
        return left.x == right.x && left.y == right.y;
    }

    public static bool operator !=( Point2 left, Point2 right ) {
        return !(left == right);
    }

    public static Point2 operator +( Point2 left, Point2 right ) {
        return new Point2( left.x + right.x, left.y + right.y );
    }

    public static Point2 operator -( Point2 left, Point2 right ) {
        return new Point2( left.x - right.x, left.y - right.y );
    }
}

// These extensions on s
public static class Point2Extensions {

    // Returns true if value is in the array
    public static bool Contains( this Point2[] points, Point2 value ) {
        foreach ( var point in points ) {
            if ( point == value ) {
                return true;
            }
        }
        return false;
    }

    // Shift all point2's in the array by a fixed offset
    public static void ShiftBy( this Point2[] points, Point2 offset ) {
        for ( int i = 0; i < points.Length; i++ ) {
            points[i] += offset;
        }
    }

    // Return a Point2 with the smallest x and y values of any point2 in the array
    // x and y need not be from the same Point2.
    public static Point2 Min( this Point2[] points ) {
        var retVal = new Point2( int.MaxValue, int.MaxValue );
        foreach ( var point in points ) {
            if ( point.x < retVal.x ) {
                retVal.x = point.x;
            }
            if ( point.y < retVal.y ) {
                retVal.y = point.y;
            }
        }
        return retVal;
    }

    // Return a Point2 with the largest x and y values of any point2 in the array
    // x and y need not be from the same Point2.
    public static Point2 Max( this Point2[] points ) {
        var retVal = new Point2( int.MinValue, int.MinValue );
        foreach ( var point in points ) {
            if ( point.x > retVal.x ) {
                retVal.x = point.x;
            }
            if ( point.y > retVal.y ) {
                retVal.y = point.y;
            }
        }
        return retVal;
    }
}
